#ifndef u_cgreeble_H
#define u_cgreeble_H

#ifdef _WINDOWS
#define WINDOWS
#endif

#include "ac_plugin.h"
#include <math.h>
#include "c3d.h"

#include <stdio.h>
#ifndef min
#define min(A,B) 	(A<B ? A : B)
#endif
#ifndef max
#define max(A,B) 	(A>B ? A : B)
#endif

//----------------------------------------------------------------------
class CGreVertex
{
public:
   // The vertex
   Vertex v;
   // The AC3D vertex
   Vertex *ac_v;
   CGreVertex()
   {
      ac_v = NULL;
   }
};
//----------------------------------------------------------------------
class CGreSVertex
{
public:
   // The vertex index
   int v;
   // TU/TV coords
   float tx, ty;
};
//----------------------------------------------------------------------
class CGreSurface
{
public:
   // Indices of vertices in the surface
   CGreSVertex *sverts;
   int numverts;
   CGreSurface()
   {
      sverts = NULL;
      numverts = 0;
   }
   ~CGreSurface()
   {
      if( sverts != NULL )
      {
         delete[] sverts;
      }
   }
};
//----------------------------------------------------------------------
class CGreObject
{

public:

   // Surfs and verts
   CGreVertex *mVerts;
   CGreSurface *mSurfs;
   unsigned int mNumVerts;
   unsigned int mNumSurfs;

   int mTexIndex;
   bool mIsCustom;

   // The bounds of the object (centered, of course)
   Vertex mMin;
   Vertex mMax;
   Vertex mCenter;

   CGreObject()
   {
      mVerts = NULL;
      mSurfs = NULL;
      mNumVerts = 0;
      mNumSurfs = 0;
      mMin.x = mMin.y = mMin.z = 0;
      mMax.x = mMax.y = mMax.z = 0;
      mCenter.x = mCenter.y = mCenter.z = 0;
      mTexIndex = -1;
      mIsCustom = true;
   }
   void ClearGreObject()
   {
      if( mVerts != NULL )
      {
         delete[] mVerts;
         mVerts = NULL;
         mNumVerts = 0;
      }
      if( mSurfs != NULL )
      {
         delete[] mSurfs;
         mSurfs = NULL;
         mNumSurfs = 0;
      }
   }
   void RotateAxis( Vertex *axis, double C )
   {
      double S;
      double a = axis->x,
             b = axis->y,
             c = axis->z;
      if( C < -1.0 )
      {
         C = -1.0;
      }
      else if( C > 1.0 )
      {
         C = 1.0;
      }
      S = sqrt( 1.0 - C * C );
      for( int ix = 0; ix < mNumVerts; ++ix )
      {
         Vertex *v = &( this->mVerts[ix].v );
         double x = v->x,
                y = v->y,
                z = v->z;
         v->x = ( float )((( 1.0 - C ) * a * a + C     ) * x + (( 1.0 - C ) * b * a - c * S ) * y + (( 1.0 - C ) * c * a + b * S ) * z );
         v->y = ( float )((( 1.0 - C ) * a * b + c * S ) * x + (( 1.0 - C ) * b * b + C     ) * y + (( 1.0 - C ) * c * b - a * S ) * z );
         v->z = ( float )((( 1.0 - C ) * a * c - b * S ) * x + (( 1.0 - C ) * b * c + a * S ) * y + (( 1.0 - C ) * c * c + C     ) * z );
      }
   }
   void CopyGreObjectFrom( CGreObject *that )
   {
      ClearGreObject();
      this->mNumSurfs = that->mNumSurfs;
      this->mNumVerts = that->mNumVerts;
      this->mVerts = new CGreVertex[this->mNumVerts];
      this->mSurfs = new CGreSurface[this->mNumSurfs];
      this->mTexIndex = that->mTexIndex;
      this->mIsCustom = that->mIsCustom;
      for( unsigned int ix = 0; ix < mNumVerts; ++ix )
      {
         this->mVerts[ix].ac_v = that->mVerts[ix].ac_v;
         this->mVerts[ix].v.x = that->mVerts[ix].v.x;
         this->mVerts[ix].v.y = that->mVerts[ix].v.y;
         this->mVerts[ix].v.z = that->mVerts[ix].v.z;
      }
      for( unsigned int ix = 0; ix < mNumSurfs; ++ix )
      {
         int numverts = that->mSurfs[ix].numverts;
         CGreSurface *s = &( this->mSurfs[ix] );
         s->numverts = numverts;
         s->sverts = new CGreSVertex[numverts];
         for( int iy = 0; iy < numverts; ++iy )
         {
            s->sverts[iy].tx = that->mSurfs[ix].sverts[iy].tx;
            s->sverts[iy].ty = that->mSurfs[ix].sverts[iy].ty;
            s->sverts[iy].v = that->mSurfs[ix].sverts[iy].v;
         }
      }
      FindBounds();
   }
   void FindBounds( bool recenter = false )
   {
      // Calculate bounds
      bool first = true;
      for( unsigned int ix = 0; ix < mNumVerts; ++ix )
      {
         Vertex *v = &( mVerts[ix].v );
         if( first )
         {
            mMin.x = v->x;
            mMin.y = v->y;
            mMin.z = v->z;
            mMax.x = v->x;
            mMax.y = v->y;
            mMax.z = v->z;
            first = false;
         }
         else
         {
            mMin.x = min( mMin.x, v->x );
            mMin.y = min( mMin.y, v->y );
            mMin.z = min( mMin.z, v->z );
            mMax.x = max( mMax.x, v->x );
            mMax.y = max( mMax.y, v->y );
            mMax.z = max( mMax.z, v->z );
         }
      }
      // Calculate center (absolute, not polycentric)
      mCenter.x = mMin.x + (( mMax.x - mMin.x ) / 2.0f );
      mCenter.y = mMin.y + (( mMax.y - mMin.y ) / 2.0f );
      mCenter.z = mMin.z + (( mMax.z - mMin.z ) / 2.0f );
      if( recenter )
      {
         Vertex bbox_offset;
         bbox_offset.x = -( mCenter.x );
         bbox_offset.y = -( mMin.y );
         bbox_offset.z = -( mCenter.z );
         for( unsigned int ix = 0; ix < mNumVerts; ++ix )
         {
            Vertex *v = &( mVerts[ix].v );
            v->x += bbox_offset.x;
            v->y += bbox_offset.y;
            v->z += bbox_offset.z;
         }
      }
   }

};
//----------------------------------------------------------------------
class CGreeble : public CGreObject
{

private:

   // Object
   ACObject *mObject;

   void PopulateVSArray();

public:

   ////////////////////////////////////////////////
   // Properties

   // Rotation increments
   float mRotationInc;
   // Scaling
   float mScaleMin;
   float mScaleMax;

   // Height of the greeble (Y axis)
   float mYheight;

   // Con/Destructor
   CGreeble();
   ~CGreeble();

   // Clear the greeble
   void ClearObject();

   // Create static greebles
   void CreateCubic();
   void CreateL();
   void CreateT();
   void CreateIII();
   // Copy an existing object.  This creates a new instance of the object
   void LoadFromObject( ACObject *o );

};
//----------------------------------------------------------------------

#endif
